using System.Text.Json;
using MediatR;
using Microsoft.AspNetCore.Authorization;
using Microsoft.AspNetCore.SignalR;
using Todo.Api.Hubs.Models;
using Todo.Core.Application.Commands.Todo;
using Todo.Core.Application.Services.UserConnectionStore;
using Todo.Core.Interfaces.Persistence;
using Todo.Core.Interfaces.User;

namespace Todo.Api.Hubs;

[Authorize]
public class TodoHub : Hub
{
    private readonly ICurrentUserService _currentUserService;
    private readonly IMediator _mediator;
    private readonly ITodoRepository _todoRepository;
    private readonly IUserConnectionStore _userConnectionStore;

    public TodoHub(
        ITodoRepository todoRepository,
        IUserConnectionStore userConnectionStore,
        ICurrentUserService currentUserService,
        IMediator mediator)
    {
        _todoRepository = todoRepository;
        _userConnectionStore = userConnectionStore;
        _currentUserService = currentUserService;
        _mediator = mediator;
    }

    public override Task OnConnectedAsync()
    {
        var userId = _currentUserService.GetUserId();
        if (userId is not null)
            _userConnectionStore.AddAsync(userId, Context.ConnectionId);
        return base.OnConnectedAsync();
    }

    public override Task OnDisconnectedAsync(Exception? exception)
    {
        var userId = _currentUserService.GetUserId();
        if (userId is not null)
            _userConnectionStore.RemoveAsync(userId, Context.ConnectionId);

        return base.OnDisconnectedAsync(exception);
    }

    public async Task CreateTodo(
        string todoTitle,
        string? projectName,
        string? description)
    {
        if (todoTitle is null)
            throw new ArgumentException("title cannot be null");

        //var userId = GetUserId();

        //var _ = await _todoRepository.CreateTodoAsync(todoTitle, projectName, userId);
        await _mediator.Send(
            new CreateTodoCommand(
                todoTitle,
                projectName,
                description));

        await GetInboxTodos();
    }


    public async Task UpdateTodo(string todoId, bool todoStatus)
    {
        var userId = GetUserId();
        await _todoRepository.UpdateTodoStatus(
            todoId,
            todoStatus,
            userId);

        await GetInboxTodos();
    }

    public async Task GetTodos()
    {
        var todos = await _todoRepository.GetTodosAsync();
        var serializedTodos = JsonSerializer.Serialize(
            todos
                .Select(
                    t => new TodoResponse
                    {
                        Id = t.Id,
                        Title = t.Title,
                        Status = t.Status,
                        Project = t.Project,
                        Description = t.Description
                    })
                .ToList());

        await RunOnUserConnections(
            async connections =>
                await Clients.Clients(connections)
                    .SendAsync("todos", serializedTodos));
    }

    public async Task GetInboxTodos()
    {
        var todos = await _todoRepository.GetNotDoneTodos();
        var serializedTodos = JsonSerializer.Serialize(
            todos
                .Select(
                    t => new TodoResponse
                    {
                        Id = t.Id,
                        Title = t.Title,
                        Status = t.Status,
                        Project = t.Project,
                        Description = t.Description
                    })
                .ToList());

        await RunOnUserConnections(
            async connections =>
                await Clients.Clients(connections)
                    .SendAsync("getInboxTodos", serializedTodos));
    }

    public async Task GetTodo(string todoId)
    {
        var todo = await _todoRepository.GetTodoByIdAsync(todoId);
        var serializedTodo = JsonSerializer.Serialize(
            new TodoResponse
            {
                Id = todo.Id,
                Project = todo.Project,
                Status = todo.Status,
                Title = todo.Title,
                Description = todo.Description
            });

        await RunOnUserConnections(
            async connections =>
                await Clients.Clients(connections)
                    .SendAsync("getTodo", serializedTodo));
    }

    public async Task ReplaceTodo(string updateTodoRequest)
    {
        var updateTodo =
            JsonSerializer.Deserialize<UpdateTodoRequest>(updateTodoRequest);
        if (updateTodo is null)
            throw new InvalidOperationException("Could not parse invalid updateTodo");

        var userId = GetUserId();

        var updatedTodo = await _todoRepository.UpdateTodoAsync(
            new Core.Entities.Todo
            {
                Id = updateTodo.Id,
                Project = updateTodo.Project,
                Status = updateTodo.Status,
                Title = updateTodo.Title,
                AuthorId = userId,
                Description = updateTodo.Description
            });

        var serializedTodo = JsonSerializer.Serialize(
            new TodoResponse
            {
                Id = updatedTodo.Id,
                Project = updatedTodo.Project,
                Status = updatedTodo.Status,
                Title = updatedTodo.Title,
                Description = updatedTodo.Description
            });

        await RunOnUserConnections(
            async connections =>
                await Clients.Clients(connections)
                    .SendAsync("getTodo", serializedTodo));
    }

    private async Task RunOnUserConnections(
        Func<IEnumerable<string>, Task> action)
    {
        var userId = GetUserId();
        var connections =
            await _userConnectionStore.GetConnectionsAsync(userId);

        await action(connections);
    }

    private string GetUserId()
    {
        return _currentUserService.GetUserId() ??
               throw new InvalidOperationException("User id was invalid. Something has gone terribly wrong");
    }
}